﻿// ReSharper disable once CheckNamespace

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.Ports;
using System.Linq;
using ServoControl.Enums;

namespace Control
{
    public class Control
    {
        private readonly object _activeLegLock = new object();
        private readonly Object _theLock = new Object();
        private List<CurrentPosition> _legDefaults;
        public LegDefaults LegFive = new LegDefaults(LegPosition.Five);
        public LegDefaults LegFour = new LegDefaults(LegPosition.Four);
        public LegDefaults LegOne = new LegDefaults(LegPosition.One);
        public LegDefaults LegSix = new LegDefaults(LegPosition.Six);
        public LegDefaults LegThree = new LegDefaults(LegPosition.Three);
        public LegDefaults LegTwo = new LegDefaults(LegPosition.Two);
        public SerialPort SerialPort { get; private set; }
        public List<CurrentPosition> LegsCurrent { get; set; }

        public List<CurrentPosition> LegDefaults
        {
            get
            {
                lock (_activeLegLock)
                    return _legDefaults;
            }
            private set
            {
                lock (_activeLegLock)
                    _legDefaults = value;
            }
        }

        public bool OverrideDefaults { get; set; }
        public int MoveSpeed { get; set; }
        public int LongDelay { get; set; }
        public int ShortDelay { get; set; }
        public bool WriteGroupMoveStrings { get; set; }

        public void Open(string comPort)
        {
            try
            {
                if (SerialPort != null)
                {
                    SerialPort.Dispose();
                    SerialPort = null;
                }

                SerialPort = new SerialPort(comPort, 38400, Parity.None, 8, StopBits.One);
                SerialPort.Open();

                SerialPort.WriteTimeout = 1000;
                SerialPort.ReadTimeout = 90000;

                SerialPort.NewLine = "\r\n";
            }
            catch (Exception e)
            {
                Debug.WriteLine(e.ToString());
            }
        }

        public void Dispose()
        {
            if (SerialPort == null) return;

            SerialPort.Dispose();
        }

        public void LoadDefaults()
        {
            var config = String.Empty;

            LegDefaults = new List<CurrentPosition>();

            LegsCurrent = new List<CurrentPosition>();

            try
            {
                config = File.ReadAllText(@"walkerDefaults.config");
            }
            catch
            {
                Debug.WriteLine(@"Cannot read walkerDefaults.config");
                return;
            }


            if (String.IsNullOrEmpty(config))
            {
                Debug.WriteLine(@"Empty config file. walkerDefaults.config");
                return;
            }

            try
            {
                var allLegDefaults = config.Split('\n');

                for (var i = 0; i != 6; i++)
                {
                    var t = allLegDefaults[i].Replace('\r'.ToString(), "");

                    var jointDefaults = t.Split('|');

                    if (i == 0)
                        LegOne = new LegDefaults(LegPosition.One)
                        {
                            CoxaServo = Convert.ToInt32(jointDefaults[0].Split(',')[2]),
                            CoxaMin = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                            CoxaMax = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                            FemurServo = Convert.ToInt32(jointDefaults[1].Split(',')[2]),
                            FemurMin = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                            FemurMax = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                            TibiaServo = Convert.ToInt32(jointDefaults[2].Split(',')[2]),
                            TibiaMin = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                            TibiaMax = Convert.ToInt32(jointDefaults[2].Split(',')[1])
                        };

                    if (i == 1)
                        LegTwo = new LegDefaults(LegPosition.Two)
                        {
                            CoxaServo = Convert.ToInt32(jointDefaults[0].Split(',')[2]),
                            CoxaMin = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                            CoxaMax = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                            FemurServo = Convert.ToInt32(jointDefaults[1].Split(',')[2]),
                            FemurMin = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                            FemurMax = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                            TibiaServo = Convert.ToInt32(jointDefaults[2].Split(',')[2]),
                            TibiaMin = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                            TibiaMax = Convert.ToInt32(jointDefaults[2].Split(',')[1])
                        };

                    if (i == 2)
                        LegThree = new LegDefaults(LegPosition.Three)
                        {
                            CoxaServo = Convert.ToInt32(jointDefaults[0].Split(',')[2]),
                            CoxaMin = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                            CoxaMax = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                            FemurServo = Convert.ToInt32(jointDefaults[1].Split(',')[2]),
                            FemurMin = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                            FemurMax = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                            TibiaServo = Convert.ToInt32(jointDefaults[2].Split(',')[2]),
                            TibiaMin = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                            TibiaMax = Convert.ToInt32(jointDefaults[2].Split(',')[1])
                        };

                    if (i == 3)
                        LegFour = new LegDefaults(LegPosition.Four)
                        {
                            CoxaServo = Convert.ToInt32(jointDefaults[0].Split(',')[2]),
                            CoxaMin = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                            CoxaMax = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                            FemurServo = Convert.ToInt32(jointDefaults[1].Split(',')[2]),
                            FemurMin = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                            FemurMax = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                            TibiaServo = Convert.ToInt32(jointDefaults[2].Split(',')[2]),
                            TibiaMin = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                            TibiaMax = Convert.ToInt32(jointDefaults[2].Split(',')[1])
                        };

                    if (i == 4)
                        LegFive = new LegDefaults(LegPosition.Five)
                        {
                            CoxaServo = Convert.ToInt32(jointDefaults[0].Split(',')[2]),
                            CoxaMin = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                            CoxaMax = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                            FemurServo = Convert.ToInt32(jointDefaults[1].Split(',')[2]),
                            FemurMin = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                            FemurMax = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                            TibiaServo = Convert.ToInt32(jointDefaults[2].Split(',')[2]),
                            TibiaMin = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                            TibiaMax = Convert.ToInt32(jointDefaults[2].Split(',')[1])
                        };

                    if (i == 5)
                        LegSix = new LegDefaults(LegPosition.Six)
                        {
                            CoxaServo = Convert.ToInt32(jointDefaults[0].Split(',')[2]),
                            CoxaMin = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                            CoxaMax = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                            FemurServo = Convert.ToInt32(jointDefaults[1].Split(',')[2]),
                            FemurMin = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                            FemurMax = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                            TibiaServo = Convert.ToInt32(jointDefaults[2].Split(',')[2]),
                            TibiaMin = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                            TibiaMax = Convert.ToInt32(jointDefaults[2].Split(',')[1])
                        };

                    LegsCurrent.Add(new CurrentPosition
                    {
                        LegPosition = (LegPosition) i,
                        Joint = Joint.Coxa,
                        Min = Convert.ToInt32(jointDefaults[0].Split(',')[0]),
                        Max = Convert.ToInt32(jointDefaults[0].Split(',')[1]),
                        Position =
                            (Convert.ToInt32(jointDefaults[0].Split(',')[0]) +
                             Convert.ToInt32(jointDefaults[0].Split(',')[1]))/2,
                        Servo = Convert.ToInt32(jointDefaults[0].Split(',')[2])
                    });

                    LegsCurrent.Add(new CurrentPosition
                    {
                        LegPosition = (LegPosition) i,
                        Joint = Joint.Femur,
                        Min = Convert.ToInt32(jointDefaults[1].Split(',')[0]),
                        Max = Convert.ToInt32(jointDefaults[1].Split(',')[1]),
                        Position =
                            (Convert.ToInt32(jointDefaults[1].Split(',')[0]) +
                             Convert.ToInt32(jointDefaults[1].Split(',')[1]))/2,
                        Servo = Convert.ToInt32(jointDefaults[1].Split(',')[2])
                    });

                    LegsCurrent.Add(new CurrentPosition
                    {
                        LegPosition = (LegPosition) i,
                        Joint = Joint.Tibia,
                        Min = Convert.ToInt32(jointDefaults[2].Split(',')[0]),
                        Max = Convert.ToInt32(jointDefaults[2].Split(',')[1]),
                        Position =
                            (Convert.ToInt32(jointDefaults[2].Split(',')[0]) +
                             Convert.ToInt32(jointDefaults[2].Split(',')[1]))/2,
                        Servo = Convert.ToInt32(jointDefaults[2].Split(',')[2])
                    });
                }
            }
            catch
            {
                //crash and burn
                return;
            }

            LegDefaults = LegsCurrent;
        }

        public void Move(int servo, int position, int speed)
        {
            if (SerialPort != null)
                SerialPort.Write("#" + servo + "P" + position + "S" + speed + Convert.ToChar(13));
        }

        public void Move(LegPosition legPosition, Joint joint, int position, int speed)
        {
            var currentLeg =
                LegsCurrent.FirstOrDefault(l => l.LegPosition == legPosition && l.Joint == joint);

            if (!OverrideDefaults && currentLeg != null)
            {
                if (position > currentLeg.Max) position = currentLeg.Max;
                if (position < currentLeg.Min) position = currentLeg.Min;
            }

            if (speed == 0) speed = 5000;

            if (currentLeg != null)
                DirectWrite("#" + currentLeg.Servo + "P" + position + "S" + speed);
        }

        public void DirectWrite(string toWrite)
        {
            if (SerialPort != null && SerialPort.IsOpen)
            {
                lock (_theLock)
                    try
                    {
                        SerialPort.Write(toWrite + Convert.ToChar(13));
                    }
                    catch (Exception e)
                    {
                        Debug.WriteLine(e.ToString());
                    }
            }
        }

        public void SendDefaultsToSsc()
        {
            var toWrite = "";

            foreach (var l in LegDefaults)
            {
                toWrite += "#" + l.Servo + "P0S" + 1000;
            }

            DirectWrite(toWrite);

            toWrite = "";

            foreach (var l in LegDefaults)
            {
                toWrite += "#" + l.Servo + "P" + l.Position + "S" + 1000;
            }

            DirectWrite(toWrite);
        }

        public void CancelCommands()
        {
            var toWrite = "";

            foreach (var l in LegDefaults)
            {
                toWrite += "#" + l.Servo + "P0S" + 1000;
            }

            DirectWrite(toWrite);
        }
    }


    public class CurrentPosition
    {
        public CurrentPosition()
        {
            LegPosition = LegPosition.One;
            Joint = Joint.Coxa;
            Position = 1000;
            Servo = 0;
        }

        public LegPosition LegPosition { get; set; }
        public Joint Joint { get; set; }
        public int Min { get; set; }
        public int Max { get; set; }
        public int Position { get; set; }
        public int Servo { get; set; }
    }

    public class LegDefaults
    {
        private int _coxaMax;
        private int _coxaMin;
        private int _femurMax;
        private int _femurMin;
        private int _tibiaMax;
        private int _tibiaMin;

        public LegDefaults(LegPosition legPosition)
        {
            LegPosition = legPosition;

            CoxaServo = 0;
            FemurServo = 0;
            TibiaServo = 0;

            CoxaMax = 0;
            CoxaMin = 0;
            CoxaCurrent = 0;
            CoxaOffset = 0;

            FemurMax = 0;
            FemurMin = 0;
            FemurCurrent = 0;
            FemurOffset = 0;

            TibiaMax = 0;
            TibiaMin = 0;
            TibiaCurrent = 0;
            TibiaOffset = 0;
        }

        public LegPosition LegPosition { get; private set; }
        public int CoxaServo { get; set; }
        public int FemurServo { get; set; }
        public int TibiaServo { get; set; }

        public int CoxaMax
        {
            get { return _coxaMax - CoxaOffset; }
            set { _coxaMax = value; }
        }

        public int CoxaMin
        {
            get { return _coxaMin + CoxaOffset; }
            set { _coxaMin = value; }
        }

        public int CoxaCurrent { get; set; }
        public int CoxaOffset { get; set; }

        public int FemurMax
        {
            get { return _femurMax - FemurOffset; }
            set { _femurMax = value; }
        }

        public int FemurMin
        {
            get { return _femurMin + FemurOffset; }
            set { _femurMin = value; }
        }

        public int FemurCurrent { get; set; }
        public int FemurOffset { get; set; }

        public int TibiaMax
        {
            get { return _tibiaMax - TibiaOffset; }
            set { _tibiaMax = value; }
        }

        public int TibiaMin
        {
            get { return _tibiaMin + TibiaOffset; }
            set { _tibiaMin = value; }
        }

        public int TibiaCurrent { get; set; }
        public int TibiaOffset { get; set; }
    }
}