﻿namespace ServoControl
{
    using System;
    using System.Collections.Generic;
    using System.Drawing;
    using System.Threading;
    using Control;

    public enum Direction
    {
        Right = 2040,
        RightMid = 1602,
        Center = 1264,
        LeftMid = 926,
        Left = 488,
        Reverse,
        ReverseTurn,
        ReverseRight,
        ReverseLeft,
        ReverseRightMid,
        ReverseLeftMid,
        Stop
    }

    public class Range
    {
        public Range()
        {
            Left = 0;
            Center = 0;
            Right = 0;
        }

        public int Left { get; set; }
        public int Center { get; set; }
        public int Right { get; set; }
    }

    public class SensorData
    {
        public SensorData()
        {
            TimeStamp = DateTime.Now;
            SensorDirection = Direction.Center;
            Range = new Range();
            Orientation = new Orientation();
        }

        public DateTime TimeStamp { get; set; }
        public Direction SensorDirection { get; set; }
        public Range Range { get; set; }
        public Orientation Orientation { get; set; }
    }

    public class Orientation
    {
        public Orientation()
        {
            Yaw = 0;
            Pitch = 0;
            Roll = 0;
        }

        public int Yaw { get; set; }
        public int Pitch { get; set; }
        public int Roll { get; set; }
    }

    public class Sensor
    {
        public readonly object MapLock = new object();
        private readonly Control _control;
        private readonly object _lookLock = new object();
        private readonly object _sensorLock = new object();
        public int CenterThreshold = 18;
        public int SideThreshold = 18;
        //private readonly List<xy> _xyfl = new List<xy>(1);
        //private readonly List<xy> _xyfr = new List<xy>(1);
        //private readonly List<xy> _xyl = new List<xy>(1);
        //private readonly List<xy> _xyr = new List<xy>(1);
        private int _lastCenter;
        private Direction _lastLookDirection;
        private int _lastSide;
        private int _lastX;
        private int _lastY;
        private Bitmap _rangeMap;
        private bool _running;
        private List<SensorData> _sensorData;
        private xy _xyPointer = new xy();
        private xy _xyc = new xy();
        private xy _xyfl = new xy();
        private xy _xyfr = new xy();
        private xy _xyl = new xy();
        private xy _xyr = new xy();
        private xy _xyrear = new xy();
        private xy _xyrl = new xy();
        private xy _xyrr = new xy();
        //private List<xy> _xyc = new List<xy>(1);

        public Sensor(Control control)
        {
            _control = control;

            LookDirection = Direction.Center;
            _lastLookDirection = Direction.Left;

            //var xy = new xy();
            //_xyc.Add(xy);

            //_xyr.Add(xy);

            //_xyl.Add(xy);

            ResetMap();
        }

        public bool WarningAverage
        {
            get
            {
                int r = (DistanceSensorCenter + DistanceSensorLeft + DistanceSensorRight)/3;
                return r < CenterThreshold;
            }
        }

        public bool WarningCenter
        {
            get { return DistanceSensorCenter < CenterThreshold; }
        }

        public bool WarningLeft
        {
            get { return DistanceSensorLeft < SideThreshold; }
        }

        public bool WarningFarLeft
        {
            get { return DistanceSensorFarLeft < SideThreshold; }
        }

        public bool WarningFarRight
        {
            get { return DistanceSensorFarRight < SideThreshold; }
        }

        public bool WarningRight
        {
            get { return DistanceSensorRight < SideThreshold; }
        }

        public bool WarningRearLeft
        {
            get { return DistanceSensorRearLeft < (SideThreshold + 5); }
        }

        public bool WarningRearRight
        {
            get { return DistanceSensorRearRight < (SideThreshold + 5); }
        }

        public bool WarningRear
        {
            get { return DistanceSensorRear < (SideThreshold + 10); }
        }

        public int DistanceSensorCenter { get; set; }
        public int DistanceSensorLeft { get; set; }
        public int DistanceSensorRight { get; set; }
        public int DistanceSensorRearLeft { get; set; }
        public int DistanceSensorRearRight { get; set; }
        public int DistanceSensorRear { get; set; }
        public double Yaw { get; set; }
        public double Pitch { get; set; }
        public double Roll { get; set; }
        public double Heading { get; set; }
        public Direction LookDirection { get; set; }
        public int DistanceSensorFarRight { get; set; }
        public int DistanceSensorFarLeft { get; set; }

        public Bitmap RangeMap
        {
            get
            {
                lock (MapLock)
                    return _rangeMap;
            }
        }

        public double HeadingDegrees
        {
            get
            {
                double degree;
                if (Yaw < 0)
                {
                    degree = Map((Math.Abs(Yaw) + 180), 180, 360, 360, 180);
                }
                else
                {
                    degree = Yaw;
                }
                return degree;
            }
        }

        public void ResetMap()
        {
            lock (MapLock)
            {
                if (_rangeMap == null)
                {
                    _rangeMap = new Bitmap(669, 506);
                }

                using (Graphics g = Graphics.FromImage(_rangeMap))
                {
                    using (Brush b = new SolidBrush(Color.Black))
                    {
                        g.FillRectangle(b, 0, 0, 669, 506);
                    }
                }
            }
        }

        public void Stop()
        {
            _running = false;
        }

        public void Start()
        {
            _running = true;

            var t = new Thread(() =>
            {
                while (_running)
                {
                    if (LookDirection != _lastLookDirection)
                    {
                        _lastLookDirection = LookDirection;
                        _control.DirectWrite("#3P" + ((int) LookDirection) + "S10000");
                        Thread.Sleep(250);
                    }

                    try
                    {
                        var strings = new string[6];
                        int tempCenter = 0;
                        int tempRight = 0;
                        int tempLeft = 0;

                        int tempFarRight = 0;
                        int tempFarLeft = 0;

                        int tempRearRight = 0;
                        int tempRearLeft = 0;

                        int tempRear = 0;

                        double tempYaw = 0;
                        double tempPitch = 0;
                        double tempRoll = 0;
                        double tempHeading = 0;

                        for (int i = 1; i <= 1; i++)
                        {
                            string data = _control.SerialPort.ReadLine();
                            strings = data.Split(':');

                            tempCenter = tempCenter + Convert.ToInt32(strings[0]);
                            tempLeft = tempLeft + Convert.ToInt32(strings[2]);
                            tempRight = tempRight + Convert.ToInt32(strings[1]);

                            tempFarLeft = tempFarLeft + Convert.ToInt32(strings[4]);
                            tempFarRight = tempFarRight + Convert.ToInt32(strings[3]);

                            tempRearLeft = tempRearLeft + Convert.ToInt32(strings[6]);
                            tempRearRight = tempRearRight + Convert.ToInt32(strings[5]);

                            tempRear = tempRear + Convert.ToInt32(strings[7]);

                            tempYaw = tempYaw + Convert.ToDouble(strings[8]);
                            tempPitch = tempPitch + Convert.ToDouble(strings[9]);
                            tempRoll = tempRoll + Convert.ToDouble(strings[10]);

                            tempHeading = tempHeading + Convert.ToDouble(strings[11]);
                        }

                        if (tempCenter > 0)
                            DistanceSensorCenter = tempCenter/1;

                        if (tempLeft > 0)
                            DistanceSensorLeft = tempLeft/1;

                        if (tempRight > 0)
                            DistanceSensorRight = tempRight/1;

                        if (tempFarLeft > 0)
                            DistanceSensorFarLeft = tempFarLeft/1;

                        if (tempFarRight > 0)
                            DistanceSensorFarRight = tempFarRight/1;

                        if (tempRearLeft > 0)
                            DistanceSensorRearLeft = tempRearLeft/1;

                        if (tempRearRight > 0)
                            DistanceSensorRearRight = tempRearRight/1;

                        if (tempRear > 0)
                            DistanceSensorRear = tempRear/1;

                        Yaw = tempYaw/3;
                        Roll = tempRoll/3;
                        Pitch = tempPitch/3;

                        if (tempHeading > 0)
                            Heading = tempHeading/3;

                        DrawMap(Yaw, DistanceSensorCenter, DistanceSensorLeft, DistanceSensorRight, DistanceSensorFarLeft, DistanceSensorFarRight, DistanceSensorRearLeft, DistanceSensorRearRight, DistanceSensorRear, Heading);
                    }
                    catch (Exception exc)
                    {
                        //Debug.WriteLine(exc.ToString());
                    }
                    Thread.Sleep(1);
                }
            }) {IsBackground = true};
            t.Start();
        }

        public static double Map(double x, double inMin, double inMax, double outMin, double outMax)
        {
            double r = (x - inMin)*(outMax - outMin)/(inMax - inMin) + outMin;
            return r;
        }

        public void DrawMap(double yaw, int range, int leftRange, int rightRange, int farLeftRange, int farRightRange, int rearLeftRange, int rearRightRange, int rearRange, double heading)
        {
            double offset = 90;

            range = range*2;
            leftRange = leftRange*2;
            rightRange = rightRange*2;

            farLeftRange = farLeftRange*2;
            farRightRange = farRightRange*2;

            farLeftRange = farLeftRange*2;
            farRightRange = farRightRange*2;

            rearLeftRange = rearLeftRange*2;
            rearRightRange = rearRightRange*2;

            rearRange = rearRange*4;

            var xyPointer = new xy();

            var xyc = new xy();
            var xyl = new xy();
            var xyr = new xy();
            var xyfl = new xy();
            var xyfr = new xy();
            var xyrl = new xy();
            var xyrr = new xy();
            var xyrear = new xy();

            double degreeC = yaw;
            double degreeFL = degreeC - 90;
            double degreeFR = degreeC + 90;

            double degreeL = degreeC - 45;
            double degreeR = degreeC + 45;

            double degreeRL = degreeC - 135;
            double degreeRR = degreeC + 135;
            double degreeRear = degreeC - 180;

            if (yaw < 0)
            {
                degreeC = (Math.Abs(yaw) + 180);
                degreeC = Map(degreeC, 180, 360, 360, 180);

                degreeFL = degreeC + 90;
                degreeFR = degreeC - 90; // * .841;

                degreeL = degreeC + 45; // * .841;
                degreeR = degreeC - 45; // * .841;

                degreeRL = degreeC + 135;
                degreeRR = degreeC - 135; // * .841;
                degreeRear = degreeC + 180;
            }

            Direction prefDirection;
            if (range > leftRange && range > rightRange && range > farRightRange && range > farLeftRange && range > rearRange && range > rearRightRange && range > rearLeftRange)
                prefDirection = Direction.Center;

            double rC = (Math.PI/180)*degreeC;
            double rL = (Math.PI/180)*degreeL;
            double rR = (Math.PI/180)*degreeR;
            double rFl = (Math.PI/180)*degreeFL;
            double rFr = (Math.PI/180)*degreeFR;
            double rRl = (Math.PI/180)*degreeRL;
            double rRr = (Math.PI/180)*degreeRR;
            double rRear = (Math.PI/180)*degreeRear;


            double xc = 315 - (Math.Cos(rC)*(range));
            double yc = 215 - (Math.Sin(rC)*(range));

            double xl = 315 - (Math.Cos(rL)*(leftRange));
            double yl = 215 - (Math.Sin(rL)*(leftRange));

            double xr = 315 - (Math.Cos(rR)*(rightRange));
            double yr = 215 - (Math.Sin(rR)*(rightRange));

            double xfl = 315 - (Math.Cos(rFl)*(farLeftRange));
            double yfl = 215 - (Math.Sin(rFl)*(farLeftRange));

            double xfr = 315 - (Math.Cos(rFr)*(farRightRange));
            double yfr = 215 - (Math.Sin(rFr)*(farRightRange));

            double xrl = 315 - (Math.Cos(rRl)*(rearLeftRange));
            double yrl = 215 - (Math.Sin(rRl)*(rearLeftRange));

            double xrr = 315 - (Math.Cos(rRr)*(rearRightRange));
            double yrr = 215 - (Math.Sin(rRr)*(rearRightRange));

            double xrear = 315 - (Math.Cos(rRear)*(rearRange));
            double yrear = 215 - (Math.Sin(rRear)*(rearRange));

            FixXy(ref xc, ref yc);
            FixXy(ref xl, ref yl);
            FixXy(ref xr, ref yr);
            FixXy(ref xfl, ref yfl);
            FixXy(ref xfr, ref yfr);
            FixXy(ref xrl, ref yrl);
            FixXy(ref xrr, ref yrr);
            FixXy(ref xrear, ref yrear);

            xyc.X = Convert.ToInt32(xc);
            xyc.Y = Convert.ToInt32(yc);

            xyl.X = Convert.ToInt32(xl);
            xyl.Y = Convert.ToInt32(yl);

            xyr.X = Convert.ToInt32(xr);
            xyr.Y = Convert.ToInt32(yr);

            xyfl.X = Convert.ToInt32(xfl);
            xyfl.Y = Convert.ToInt32(yfl);

            xyfr.X = Convert.ToInt32(xfr);
            xyfr.Y = Convert.ToInt32(yfr);

            xyrl.X = Convert.ToInt32(xrl);
            xyrl.Y = Convert.ToInt32(yrl);

            xyrr.X = Convert.ToInt32(xrr);
            xyrr.Y = Convert.ToInt32(yrr);

            xyrear.X = Convert.ToInt32(xrear);
            xyrear.Y = Convert.ToInt32(yrear);

            int averageX = (xyc.X + xyr.X + xyl.X)/3;
            int averageY = (xyc.Y + xyr.Y + xyl.Y)/3;

            lock (MapLock)
            {
                if (_rangeMap == null)
                {
                    _rangeMap = new Bitmap(669, 506);
                }


                using (Graphics g = Graphics.FromImage(_rangeMap))
                {
                    using (Brush b = new SolidBrush(Color.Black))
                    {
                        g.FillRectangle(b, 0, 0, 669, 506);
                    }


                    g.DrawLine(new Pen(Brushes.DarkRed, 2), 315, 215, xyc.X, xyc.Y);


                    var points = new Point[9];
                    points[0] = new Point(xyc.X, xyc.Y);
                    points[1] = new Point(xyr.X, xyr.Y);
                    points[2] = new Point(xyfr.X, xyfr.Y);
                    points[3] = new Point(xyrr.X, xyrr.Y);
                    points[4] = new Point(xyrear.X, xyrear.Y);
                    points[5] = new Point(xyrl.X, xyrl.Y);
                    points[6] = new Point(xyfl.X, xyfl.Y);
                    points[7] = new Point(xyl.X, xyl.Y);
                    points[8] = new Point(xyc.X, xyc.Y);

                    g.DrawCurve(new Pen(Color.White), points);

                    _lastCenter = CenterThreshold;

                    _lastSide = SideThreshold;

                    _lastX = averageX;
                    _lastY = averageY;

                    _xyPointer = xyPointer;

                    _xyc = xyc;
                    _xyr = xyr;
                    _xyl = xyl;
                    _xyfr = xyfr;
                    _xyfl = xyfl;
                    _xyrr = xyrr;
                    _xyrl = xyrl;
                    _xyrear = xyrear;
                }
            }

            Thread.Sleep(1);
        }

        private void FixXy(ref double x, ref double y)
        {
            if (x > 669) x = 669;
            if (y > 506) y = 506;
            if (x < 0) x = 0;
            if (y < 0) y = 0;
        }
    }

    internal class xy
    {
        public xy()
        {
            X = 315;
            Y = 215;
            Age = DateTime.Now;
        }

        public int X { get; set; }
        public int Y { get; set; }
        public DateTime Age { get; set; }
    }
}