﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;
using ServoControl;
using ServoControl.Enums;

namespace Control
{
    // ReSharper disable once IdentifierTypo
    public partial class HexConfigurator : Form
    {
        private readonly InverseKinematics _inverseKinematics;
        private int _coxaAngle = 0;
        private int _femuAngle = 0;
        private bool _leftButton;
        private bool _listening;
        private bool _middleButton;
        private bool _rightButton;
        private List<Graphics> _shapes = new List<Graphics>();
        private int _tibiaAngle = 0;

        public HexConfigurator()
        {
            InitializeComponent();

            GetComPorts();

            lbLeg.SelectedIndex = 0;

            FormClosing += HexConfigurator_FormClosing;

            _control = new Control();
            _sensor = new Sensor(_control);
            _inverseKinematics = new InverseKinematics(_control, _sensor);
            _automation = new Automation(_control, _sensor, _inverseKinematics);

            gaitType.SelectedIndex = _inverseKinematics.GaitType;

            LegLiftHeight.Value = Convert.ToInt32(_inverseKinematics.LegLiftHeight);
            LegLiftHeight.Text = _inverseKinematics.LegLiftHeight.ToString(CultureInfo.InvariantCulture);

            BodyPositionY.Value = Convert.ToInt32(_inverseKinematics.BodyPosY);
            BodyPosYTxt.Text = _inverseKinematics.BodyPosY.ToString(CultureInfo.InvariantCulture);

            LoadDefaults();

            MouseWheel += HexConfigurator_MouseWheel;
            RemoteControlPanel.MouseMove += RemoteControlPanel_MouseMove;
            RemoteControlPanel.MouseClick += HexConfigurator_MouseClick;

            RotatePanel.MouseMove += RotatePanel_MouseMove;
            RotatePanel.MouseClick += HexConfigurator_MouseClick;
        }

        private void RotatePanel_MouseMove(object sender, MouseEventArgs e)
        {
            if (_leftButton && _enableRemote)
            {
                if (e.Y < 50)
                {
                    var val = Convert.ToDouble(-Math.Abs((e.Y - 50)/2));

                    if (val < -15) val = -15;

                    _inverseKinematics.BodyRotZ1 = val;
                    BodyRotZ.Value = Convert.ToInt32(val);
                    BodyRotZtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }

                if (e.Y > 50)
                {
                    var val = Convert.ToDouble(Math.Abs((e.Y - 50)/2));

                    if (val > 15) val = 15;

                    _inverseKinematics.BodyRotZ1 = val;
                    BodyRotZ.Value = Convert.ToInt32(val);
                    BodyRotZtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }

                if (e.X < 50)
                {
                    var val = Convert.ToDouble(-Math.Abs((((e.X - 50)/2))));

                    if (val < -15) val = -15;

                    _inverseKinematics.BodyRotX1 = val;
                    BodyRotX.Value = Convert.ToInt32(val);
                    BodyRotXtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }

                if (e.X > 50)
                {
                    var val = Convert.ToDouble(Math.Abs((((e.X - 50)/2))));

                    if (val > 15) val = 15;

                    _inverseKinematics.BodyRotX1 = val;
                    BodyRotX.Value = Convert.ToInt32(val);
                    BodyRotXtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }
            }
        }

        private void HexConfigurator_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (_leftButton)
                {
                    _leftButton = false;
                }
                else
                    _leftButton = true;
            }

            if (e.Button == MouseButtons.Right)
            {
                if (_rightButton)
                {
                    _rightButton = false;
                }
                else
                    _rightButton = true;
            }

            if (e.Button == MouseButtons.Middle)
            {
                if (_middleButton)
                {
                    _middleButton = false;
                }
                else
                    _middleButton = true;
            }
        }

        private void RemoteControlPanel_MouseMove(object sender, MouseEventArgs e)
        {
            if (_leftButton && _enableRemote)
            {
                if (e.Y < 125)
                {
                    var val = Convert.ToDouble(-Math.Abs((e.Y - 125)/2));

                    _inverseKinematics.TravelLengthZ = val;
                    TravelLengthZ.Value = Convert.ToInt32(val);
                    travelZtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }
                if (e.Y > 125)
                {
                    var val = Convert.ToDouble(Math.Abs((e.Y - 125)/2));

                    _inverseKinematics.TravelLengthZ = val;
                    TravelLengthZ.Value = Convert.ToInt32(val);
                    travelZtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }

                if (e.X < 125)
                {
                    var val = Convert.ToDouble(-Math.Abs((((e.X - 125)/20))));

                    if (val > 3) val = 3;
                    if (val < -3) val = -3;

                    _inverseKinematics.TravelRotationY = val;
                    TravelRotY.Value = Convert.ToInt32(val);
                    TravelRotYtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }

                if (e.X > 125)
                {
                    var val = Convert.ToDouble(Math.Abs((((e.X - 125)/20))));

                    if (val > 3) val = 3;
                    if (val < -3) val = -3;

                    _inverseKinematics.TravelRotationY = val;
                    TravelRotY.Value = Convert.ToInt32(val);
                    TravelRotYtxt.Text = val.ToString(CultureInfo.InvariantCulture);
                }
            }
        }

        private void HexConfigurator_MouseWheel(object sender, MouseEventArgs e)
        {
            if (!_enableRemote)
                return;

            if (_rightButton)
            {
                var val = e.Delta;

                var newVal = 0;

                if (val < 0)
                {
                    newVal = LegLiftHeight.Value - 5;
                    if (val < 0) val = 0;
                }

                if (val > 0)
                {
                    newVal = LegLiftHeight.Value + 5;
                    if (newVal > 110) newVal = 110;
                }

                LegLiftHeight.Value = newVal;
                LegLiftHeighttxt.Text = newVal.ToString();
                _inverseKinematics.LegLiftHeight = newVal;
                return;
            }

            if (_leftButton)
            {
                var newVal2 = 0;

                if (e.Delta < 0)
                {
                    newVal2 = BodyPositionY.Value - 5;
                    if (newVal2 < 0) newVal2 = 0;
                }

                if (e.Delta > 0)
                {
                    newVal2 = BodyPositionY.Value + 5;
                    if (newVal2 > 110) newVal2 = 110;
                }

                BodyPositionY.Value = newVal2;
                BodyPosYTxt.Text = newVal2.ToString();
                _inverseKinematics.BodyPosY = newVal2;
            }
        }

        private void HexConfigurator_FormClosing(object sender, FormClosingEventArgs e)
        {
            _control.CancelCommands();

            Thread.Sleep(500);

            _control.Dispose();
        }

        System.Timers.Timer _updateTimer = new System.Timers.Timer(1000);

        private void LoadDefaults()
        {
            if (!File.Exists(Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + @"\walkerDefaults.config"))
            {
                MessageBox.Show(@"walkerDefaults.confg does not exist in " + Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location), @"Ok", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            _control.LoadDefaults();
            dgDefaults.DataSource = _control.LegDefaults;
        }

        private void lbLeg_MouseClick(object sender, MouseEventArgs e)
        {
            ActiveLegPosition = (LegPosition) lbLeg.SelectedIndex;
        }

        private void rbCoxa_MouseClick(object sender, MouseEventArgs e)
        {
            ActiveJoint = ServoControl.Enums.Joint.Coxa;
        }

        private void rbFemur_MouseClick(object sender, MouseEventArgs e)
        {
            ActiveJoint = ServoControl.Enums.Joint.Femur;
        }

        private void rbTibia_MouseClick(object sender, MouseEventArgs e)
        {
            ActiveJoint = ServoControl.Enums.Joint.Tibia;
        }

        private void overrideSafeValues_CheckedChanged(object sender, EventArgs e)
        {
            _control.OverrideDefaults = overrideSafeValues.Checked;
        }

        private void AuxPosition_Scroll(object sender, ScrollEventArgs e)
        {
            _sweepCurrent = e.NewValue;
            _control.DirectWrite("#" + Convert.ToInt32(ServoNumber.Text) + "P" + e.NewValue + "S20000");
        }

        private void lbLeg_SelectedIndexChanged(object sender, EventArgs e)
        {
            ActiveLegPosition = (LegPosition) lbLeg.SelectedIndex;
        }

        private void RemoteControl_CheckedChanged(object sender, EventArgs e)
        {
            _enableRemote = RemoteControl.Checked;
        }

        private void HexConfigurator_KeyUp(object sender, KeyEventArgs e)
        {
            if (_enableRemote)
            {
                switch (e.KeyCode)
                {
                    case Keys.Up:
                        _inverseKinematics.TravelLengthZ = _inverseKinematics.TravelLengthZ - 15;
                        TravelLengthZ.Value = Convert.ToInt32(_inverseKinematics.TravelLengthZ);
                        break;
                    case Keys.Down:
                        _inverseKinematics.TravelLengthZ = _inverseKinematics.TravelLengthZ + 15;
                        TravelLengthZ.Value = Convert.ToInt32(_inverseKinematics.TravelLengthZ);
                        break;
                    case Keys.Right:
                        _inverseKinematics.TravelRotationY = _inverseKinematics.TravelRotationY + 1;
                        TravelRotY.Value = Convert.ToInt32(_inverseKinematics.TravelRotationY);
                        break;
                    case Keys.Left:
                        _inverseKinematics.TravelRotationY = _inverseKinematics.TravelRotationY - 1;
                        TravelRotY.Value = Convert.ToInt32(_inverseKinematics.TravelRotationY);
                        break;
                }
            }
        }

        private void UltraSweep_CheckedChanged(object sender, EventArgs e)
        {
            _sweepMin = AuxPosition.Minimum;
            _sweepMax = AuxPosition.Maximum;
            _sweepCurrent = _sweepMin;

            _sweepRunning = UltraSweep.Checked;

            if (UltraSweep.Checked)
            {
                _sweepThread = new Thread(() =>
                {
                    while (_sweepRunning)
                    {
                        Thread.Sleep(1);
                        if (_sweepCurrent > (_sweepMax - 7))
                        {
                            _sweepDirection = false;
                            _control.Move(Convert.ToInt32(ServoNumber.Text), 688, 615);
                        }

                        if (_sweepCurrent < (_sweepMin + 7))
                        {
                            _sweepDirection = true;
                            _control.Move(Convert.ToInt32(ServoNumber.Text), 1870, 615);
                        }

                        if (_sweepDirection)
                            _sweepCurrent = _sweepCurrent + 7;
                        else
                            _sweepCurrent = _sweepCurrent - 7;

                        if (!Disposing && !IsDisposed)
                        {
                            Invoke((MethodInvoker) delegate { AuxPosition.Value = _sweepCurrent; });
                        }

                        Thread.Sleep(10);
                    }
                    ;
                });
                _sweepThread.IsBackground = true;
                _sweepThread.Start();
            }
        }

        private void HexConfigurator_FormClosed(object sender, FormClosedEventArgs e)
        {
            _control.CancelCommands();
            _control.Dispose();
        }

        private void SetMidPoint_Click(object sender, EventArgs e)
        {
            var offset = (Convert.ToInt32(servoRange.Text)/2);

            switch (ActiveJoint)
            {
                case ServoControl.Enums.Joint.Coxa:
                    dgDefaults.SelectedRows[0].Cells[3].Value = Convert.ToInt32(waistCurrent.Text) + offset;
                    dgDefaults.SelectedRows[0].Cells[2].Value = Convert.ToInt32(waistCurrent.Text) - offset;
                    break;
                case ServoControl.Enums.Joint.Femur:
                    dgDefaults.SelectedRows[0].Cells[3].Value = Convert.ToInt32(hipCurrent.Text) + offset;
                    dgDefaults.SelectedRows[0].Cells[2].Value = Convert.ToInt32(hipCurrent.Text) - offset;
                    break;
                case ServoControl.Enums.Joint.Tibia:
                    dgDefaults.SelectedRows[0].Cells[3].Value = Convert.ToInt32(kneeCurrent.Text) + offset;
                    dgDefaults.SelectedRows[0].Cells[2].Value = Convert.ToInt32(kneeCurrent.Text) - offset;
                    break;
            }
        }

        // ReSharper disable once FunctionComplexityOverflow
        private void SetSelectedDgRow()
        {
            if (dgDefaults.Rows.Count > 0)
                switch (_activeLegPosition)
                {
                    case LegPosition.One:
                        dgDefaults.Rows[(int) ActiveJoint].Selected = true;

                        Invoke((MethodInvoker) delegate
                        {
                            waistPosition.Value = Convert.ToInt32(dgDefaults.Rows[0].Cells[2].Value) +
                                                  ((Convert.ToInt32(dgDefaults.Rows[0].Cells[3].Value) -
                                                    Convert.ToInt32(dgDefaults.Rows[0].Cells[2].Value))/2);
                            waistCurrent.Text =
                                Convert.ToString(Convert.ToInt32(dgDefaults.Rows[0].Cells[2].Value) +
                                                 ((Convert.ToInt32(dgDefaults.Rows[0].Cells[3].Value) -
                                                   Convert.ToInt32(dgDefaults.Rows[0].Cells[2].Value))/2));


                            hipPosition.Value = Convert.ToInt32(dgDefaults.Rows[1].Cells[2].Value) +
                                                ((Convert.ToInt32(dgDefaults.Rows[1].Cells[3].Value) -
                                                  Convert.ToInt32(dgDefaults.Rows[1].Cells[2].Value))/2);
                            hipCurrent.Text =
                                Convert.ToString(Convert.ToInt32(dgDefaults.Rows[1].Cells[2].Value) +
                                                 ((Convert.ToInt32(dgDefaults.Rows[1].Cells[3].Value) -
                                                   Convert.ToInt32(dgDefaults.Rows[1].Cells[2].Value))/2));

                            kneePosition.Value = Convert.ToInt32(dgDefaults.Rows[2].Cells[2].Value) +
                                                 ((Convert.ToInt32(dgDefaults.Rows[2].Cells[3].Value) -
                                                   Convert.ToInt32(dgDefaults.Rows[2].Cells[2].Value))/2);
                            kneeCurrent.Text =
                                Convert.ToString(Convert.ToInt32(dgDefaults.Rows[2].Cells[2].Value) +
                                                 ((Convert.ToInt32(dgDefaults.Rows[2].Cells[3].Value) -
                                                   Convert.ToInt32(dgDefaults.Rows[2].Cells[2].Value))/2));
                        });
                        break;
                    case LegPosition.Two:
                        dgDefaults.Rows[(int) ActiveJoint + 3].Selected = true;

                        Invoke((MethodInvoker) delegate
                        {
                            waistPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[3].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[3].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[3].Cells[2].Value))/2);
                            waistCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[3].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[3].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[3].Cells[2].Value))/2));

                            hipPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[4].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[4].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[4].Cells[2].Value))/2);
                            hipCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[4].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[4].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[4].Cells[2].Value))/2));

                            kneePosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[5].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[5].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[5].Cells[2].Value))/2);
                            kneeCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[5].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[5].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[5].Cells[2].Value))/2));
                        });
                        break;
                    case LegPosition.Three:
                        dgDefaults.Rows[(int) ActiveJoint + 6].Selected = true;

                        Invoke((MethodInvoker) delegate
                        {
                            waistPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[6].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[6].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[6].Cells[2].Value))/2);
                            waistCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[6].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[6].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[6].Cells[2].Value))/2));

                            hipPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[7].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[7].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[7].Cells[2].Value))/2);
                            hipCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[7].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[7].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[7].Cells[2].Value))/2));

                            kneePosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[8].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[8].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[8].Cells[2].Value))/2);
                            kneeCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[8].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[8].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[8].Cells[2].Value))/2));
                        });
                        break;
                    case LegPosition.Four:
                        dgDefaults.Rows[(int) ActiveJoint + 9].Selected = true;

                        Invoke((MethodInvoker) delegate
                        {
                            waistPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[9].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[9].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[9].Cells[2].Value))/2);
                            waistCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[9].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[9].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[9].Cells[2].Value))/2));

                            hipPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[10].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[10].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[10].Cells[2].Value))/2);
                            hipCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[10].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[10].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[10].Cells[2].Value))/2));

                            kneePosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[11].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[11].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[11].Cells[2].Value))/2);
                            kneeCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[11].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[11].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[11].Cells[2].Value))/2));
                        });
                        break;
                    case LegPosition.Five:
                        dgDefaults.Rows[(int) ActiveJoint + 12].Selected = true;

                        Invoke((MethodInvoker) delegate
                        {
                            waistPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[12].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[12].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[12].Cells[2].Value))/2);
                            waistCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[12].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[12].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[12].Cells[2].Value))/2));

                            hipPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[13].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[13].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[13].Cells[2].Value))/2);
                            hipCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[13].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[13].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[13].Cells[2].Value))/2));

                            kneePosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[14].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[14].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[14].Cells[2].Value))/2);
                            kneeCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[14].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[14].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[14].Cells[2].Value))/2));
                        });
                        break;
                    case LegPosition.Six:
                        dgDefaults.Rows[(int) ActiveJoint + 15].Selected = true;

                        Invoke((MethodInvoker) delegate
                        {
                            waistPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[15].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[15].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[15].Cells[2].Value))/2);
                            waistCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[15].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[15].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[15].Cells[2].Value))/2));

                            hipPosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[16].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[16].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[16].Cells[2].Value))/2);
                            hipCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[16].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[16].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[16].Cells[2].Value))/2));

                            kneePosition.Value =
                                Convert.ToInt32(dgDefaults.Rows[17].Cells[2].Value) +
                                ((Convert.ToInt32(dgDefaults.Rows[17].Cells[3].Value) -
                                  Convert.ToInt32(dgDefaults.Rows[17].Cells[2].Value))/2);
                            kneeCurrent.Text =
                                Convert.ToString(
                                    Convert.ToInt32(dgDefaults.Rows[17].Cells[2].Value) +
                                    ((Convert.ToInt32(dgDefaults.Rows[17].Cells[3].Value) -
                                      Convert.ToInt32(dgDefaults.Rows[17].Cells[2].Value))/2));
                        });
                        break;
                }
        }

        private void GetComPorts()
        {
            var ports = SerialPort.GetPortNames();

            foreach (var p in ports)
            {
                ComPortList.Items.Add(p);
            }
        }

        private void TravelLengthX_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.TravelLengthX = Convert.ToDouble(e.NewValue);

            travelXtxt.Text = (Convert.ToDouble(e.NewValue)).ToString(CultureInfo.InvariantCulture);
        }

        private void TravelLengthZ_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.TravelLengthZ = Convert.ToDouble(e.NewValue);

            travelZtxt.Text = Convert.ToDouble(e.NewValue).ToString(CultureInfo.InvariantCulture);
        }

        private void BodyPositionX_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.BodyPosX = Convert.ToDouble(e.NewValue);

            bodyPosXTxt.Text = e.NewValue.ToString();
        }

        private void BodyPositionY_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.BodyPosY = Convert.ToDouble(e.NewValue);

            BodyPosYTxt.Text = e.NewValue.ToString();
        }

        private void BodyPositionZ_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.BodyPosZ = Convert.ToDouble(e.NewValue)/4;

            bodyPosZTxt.Text = Convert.ToDouble(e.NewValue/4).ToString(CultureInfo.InvariantCulture);
        }

        private void comboBox3_SelectedIndexChanged(object sender, EventArgs e)
        {
            _inverseKinematics.GaitType = ((ComboBox) sender).SelectedIndex;

            _inverseKinematics.GaitSelect();
        }

        private void OneK_Click(object sender, EventArgs e)
        {
            gaitType.SelectedIndex = 0;
        }

        private void StartStop1k_Click_1(object sender, EventArgs e)
        {
            if (_inverseKinematics.Running)
            {
                _inverseKinematics.Running = false;
                _control.CancelCommands();
            }
            else
            {
                _inverseKinematics.Setup();
                _inverseKinematics.Running = true;
                _inverseKinematics.Run(UpdateDebugTxt);

                BodyPosYTxt.Text = _inverseKinematics.BodyPosY.ToString(CultureInfo.InvariantCulture);
                BodyPositionY.Value = Convert.ToInt32(_inverseKinematics.BodyPosY);

                travelZtxt.Text = _inverseKinematics.TravelLengthZ.ToString(CultureInfo.InvariantCulture);
                BodyPositionZ.Value = Convert.ToInt32(_inverseKinematics.BodyPosZ);

                BodyRotXtxt.Text = _inverseKinematics.BodyRotX1.ToString(CultureInfo.InvariantCulture);
                BodyRotX.Value = Convert.ToInt32(_inverseKinematics.BodyRotX1);

                BodyRotZtxt.Text = _inverseKinematics.BodyRotZ1.ToString(CultureInfo.InvariantCulture);
                BodyRotZ.Value = Convert.ToInt32(_inverseKinematics.BodyRotZ1);

                TravelRotYtxt.Text = _inverseKinematics.TravelRotationY.ToString(CultureInfo.InvariantCulture);
                TravelRotY.Value = Convert.ToInt32(_inverseKinematics.TravelRotationY);

                NomGaitSpeed.Value = 150;
                _inverseKinematics.NomGaitSpeed = 150;
                NomGaitSpeedTxt.Text = "150";

                LegLiftHeight.Value = 10;
                LegLiftHeighttxt.Text = "10";

                TravelLengthX.Value = 0;
                travelXtxt.Text = "0";

                bodyPosXTxt.Text = "0";
                BodyPositionX.Value = 0;

                BodyPositionZ.Value = 0;
                bodyPosZTxt.Text = "0";
            }
        }

        private void UpdateDebugTxt(string dbg)
        {
            try
            {
                

                Invoke((MethodInvoker) delegate
                {
                    
                    debugTxt.Text = dbg;


                });
            }
            catch (ObjectDisposedException)
            {
                //Disposed...
            }
        }

        private void TravelRotY_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.TravelRotationY = Convert.ToDouble(e.NewValue);

            TravelRotYtxt.Text = _inverseKinematics.TravelRotationY.ToString(CultureInfo.InvariantCulture);
        }

        private void LegLiftHeight_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.LegLiftHeight = e.NewValue;
            LegLiftHeighttxt.Text = e.NewValue.ToString();
        }

        private void OpenCloseBtn_Click(object sender, EventArgs e)
        {
            if (OpenCloseBtn.Text.Contains("Close"))
            {
                OpenCloseBtn.Text = @"Open";
                _control.SerialPort.Dispose();
                return;
            }

            if (!File.Exists(Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + @"\walkerDefaults.config"))
            {
                MessageBox.Show(@"walkerDefaults.config does not exist in " + Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location), @"Ok", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            if (ComPortList.Items.Count == 0 || string.IsNullOrEmpty(ComPortList.SelectedItem.ToString()))
                return;

            _control.Open(ComPortList.SelectedItem.ToString());

            if (_control.SerialPort.IsOpen)
            {
                _sensor.Start();

                //_updateTimer.Elapsed += _updateTimer_Elapsed;
                //_updateTimer.Start();

                var t = new Thread(() =>
                {
                    while (true)
                    {
                        Thread.Sleep(1);
                        try
                        {
                            Invoke((MethodInvoker)delegate
                            {
                                DistanceCenter.Text = _sensor.DistanceSensorCenter.ToString();
                                DistanceLeft.Text = _sensor.DistanceSensorLeft.ToString();
                                DistanceRight.Text = _sensor.DistanceSensorRight.ToString();

                                DistanceSideLeft.Text = _sensor.DistanceSensorFarLeft.ToString();
                                DistanceSideRight.Text = _sensor.DistanceSensorFarLeft.ToString();

                                DistanceRearLeft.Text = _sensor.DistanceSensorRearLeft.ToString();
                                DistanceRearRight.Text = _sensor.DistanceSensorRearRight.ToString();
                                DistanceRearCenter.Text = _sensor.DistanceSensorRear.ToString();

                                Yaw.Text = _sensor.Yaw.ToString();
                                Pitch.Text = _sensor.Pitch.ToString();
                                Roll.Text = _sensor.Roll.ToString();
                                Heading.Text = _sensor.Heading.ToString();


                                if (_sensor.RangeMap != null)
                                {
                                    lock (_sensor.MapLock)
                                        using (var graphics = MapPanel.CreateGraphics())
                                            graphics.DrawImage(_sensor.RangeMap, 0, 0);                                    

                                }

                            });
                        }
                        catch (Exception)
                        {
                        }

                    }
                });
                t.IsBackground = true;
                t.Start();

                OpenCloseBtn.Text = @"Close";
                //MessageBox.Show(ComPortList.SelectedItem + @" opened", @"Ok", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                _sensor.Stop();

                OpenCloseBtn.Text = @"Open";
               // MessageBox.Show(@"Open " + ComPortList.SelectedItem + @" Failed", @"Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        void _updateTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            //MapPanel.Invalidate();

            //lock (_sensor.MapLock)
            //using (var graphics = MapPanel.CreateGraphics())
            //    graphics.DrawImage(_sensor.RangeMap, 0, 0);
        }

        private void LoadDefaultsBtn_Click(object sender, EventArgs e)
        {
            LoadDefaults();
        }

        private void SaveDefaultsBtn_Click(object sender, EventArgs e)
        {
            var toWrite = "";
            var count = 0;

            foreach (DataGridViewRow r in dgDefaults.Rows)
            {
                if ((LegPosition) r.Cells[0].Value == LegPosition.One)
                {
                    toWrite += r.Cells[2].Value + "," + r.Cells[3].Value + "," + r.Cells[7].Value + "," +
                               r.Cells[6].Value + "," + r.Cells[5].Value;
                    if (count == 2)
                    {
                        toWrite += "\n";
                        count = 0;
                    }
                    else
                    {
                        toWrite += "|";
                        count++;
                    }
                    continue;
                }

                if ((LegPosition) r.Cells[0].Value == LegPosition.Two)
                {
                    toWrite += r.Cells[2].Value + "," + r.Cells[3].Value + "," + r.Cells[7].Value + "," +
                               r.Cells[6].Value + "," + r.Cells[5].Value;
                    if (count == 2)
                    {
                        toWrite += "\n";
                        count = 0;
                    }
                    else
                    {
                        toWrite += "|";
                        count++;
                    }
                    continue;
                }

                if ((LegPosition) r.Cells[0].Value == LegPosition.Three)
                {
                    toWrite += r.Cells[2].Value + "," + r.Cells[3].Value + "," + r.Cells[7].Value + "," +
                               r.Cells[6].Value + "," + r.Cells[5].Value;
                    if (count == 2)
                    {
                        toWrite += "\n";
                        count = 0;
                    }
                    else
                    {
                        toWrite += "|";
                        count++;
                    }
                    continue;
                }

                if ((LegPosition) r.Cells[0].Value == LegPosition.Four)
                {
                    toWrite += r.Cells[2].Value + "," + r.Cells[3].Value + "," + r.Cells[7].Value + "," +
                               r.Cells[6].Value + "," + r.Cells[5].Value;
                    if (count == 2)
                    {
                        toWrite += "\n";
                        count = 0;
                    }
                    else
                    {
                        toWrite += "|";
                        count++;
                    }
                    continue;
                }

                if ((LegPosition) r.Cells[0].Value == LegPosition.Five)
                {
                    toWrite += r.Cells[2].Value + "," + r.Cells[3].Value + "," + r.Cells[7].Value + "," +
                               r.Cells[6].Value + "," + r.Cells[5].Value;
                    if (count == 2)
                    {
                        toWrite += "\n";
                        count = 0;
                    }
                    else
                    {
                        toWrite += "|";
                        count++;
                    }
                    continue;
                }

                if ((LegPosition) r.Cells[0].Value == LegPosition.Six)
                {
                    toWrite += r.Cells[2].Value + "," + r.Cells[3].Value + "," + r.Cells[7].Value + "," +
                               r.Cells[6].Value + "," + r.Cells[5].Value;
                    if (count == 2)
                    {
                        toWrite += "\n";
                        count = 0;
                    }
                    else
                    {
                        toWrite += "|";
                        count++;
                    }
                }
            }

            try
            {
                File.WriteAllText(@"walkerDefaults.config", toWrite);
                MessageBox.Show(@"Config saved to walkerDefaults.config", @"Ok", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch
            {
                MessageBox.Show(@"Could not save config file", @"Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void InitializeServosBtn_Click(object sender, EventArgs e)
        {
            _control.CancelCommands();
        }

        private void HomeServosBtn_Click(object sender, EventArgs e)
        {
            LoadDefaults();
            _control.SendDefaultsToSsc();
        }

        private void NomGaitSpeed_Scroll(object sender, ScrollEventArgs e)
        {
            NomGaitSpeedTxt.Text = e.NewValue.ToString();
            _inverseKinematics.NomGaitSpeed = e.NewValue;
        }

        private void BodyRotZ_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.BodyRotZ1 = e.NewValue;

            BodyRotZtxt.Text = e.NewValue.ToString();
        }

        private void BodyRotX_Scroll(object sender, ScrollEventArgs e)
        {
            _inverseKinematics.BodyRotX1 = e.NewValue;

            BodyRotXtxt.Text = e.NewValue.ToString();
        }

        #region Properties

        private readonly Bitmap _buffer = new Bitmap(40, 5);

        private readonly Bitmap _legBuffer = new Bitmap(100, 100);

        private Joint _activeJoint;
        private LegPosition _activeLegPosition;

        private bool _enableRemote;
        private int _sweepCurrent = 1;
        private bool _sweepDirection;
        private int _sweepMax = 1;
        private int _sweepMin = 1;
        private bool _sweepRunning;
        private Thread _sweepThread;
        private readonly int _y = 10; //vert
        private Control _control;
        private Sensor _sensor;
        private Automation _automation;

        private LegPosition ActiveLegPosition
        {
            set
            {
                _activeLegPosition = value;

                SetSelectedDgRow();

                lbLeg.SelectedIndex = (int) _activeLegPosition;
            }
            get { return _activeLegPosition; }
        }

        private Joint ActiveJoint
        {
            set
            {
                _activeJoint = value;

                SetSelectedDgRow(); //dont go setting active joint willynilly

                if (_activeJoint == ServoControl.Enums.Joint.Coxa)
                {
                    rbCoxa.Checked = true;
                    rbFemur.Checked = false;
                    rbTibia.Checked = false;

                    switch (ActiveLegPosition)
                    {
                        case LegPosition.One:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Two:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 2].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 2].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Three:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 4].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 4].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Four:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 6].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 6].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Five:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 8].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 8].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Six:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 10].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 10].Cells[2].Value)).ToString
                                    ();
                            break;
                    }
                }
                if (_activeJoint == ServoControl.Enums.Joint.Femur)
                {
                    rbCoxa.Checked = false;
                    rbFemur.Checked = true;
                    rbTibia.Checked = false;

                    switch (ActiveLegPosition)
                    {
                        case LegPosition.One:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 1].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 1].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Two:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 3].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 3].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Three:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 5].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 5].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Four:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 7].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 7].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Five:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 9].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 9].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Six:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 11].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 11].Cells[2].Value)).ToString
                                    ();
                            break;
                    }
                }

                if (_activeJoint == ServoControl.Enums.Joint.Tibia)
                {
                    rbCoxa.Checked = false;
                    rbFemur.Checked = false;
                    rbTibia.Checked = true;

                    switch (ActiveLegPosition)
                    {
                        case LegPosition.One:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 2].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 2].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Two:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 4].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 4].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Three:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 6].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 6].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Four:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 8].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 8].Cells[2].Value)).ToString();
                            break;
                        case LegPosition.Five:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 10].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 10].Cells[2].Value)).ToString
                                    ();
                            break;
                        case LegPosition.Six:
                            servoRange.Text =
                                (Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 12].Cells[3].Value) -
                                 Convert.ToInt32(dgDefaults.Rows[(int) ActiveLegPosition + 12].Cells[2].Value)).ToString
                                    ();
                            break;
                    }
                }
            }
            get { return _activeJoint; }
        }

        #endregion

        #region Position

        private void waistPosition_Scroll(object sender, ScrollEventArgs e)
        {
            Invoke((MethodInvoker) delegate
            {
                var t = e.NewValue; // + Convert.ToInt32(ssc.activeMin);

                ActiveJoint = ServoControl.Enums.Joint.Coxa;

                _control.Move(ActiveLegPosition, ServoControl.Enums.Joint.Coxa, t, 5000);

                var offset = 0;

                if (ActiveLegPosition == LegPosition.Two) offset = 2;
                if (ActiveLegPosition == LegPosition.Three) offset = 4;
                if (ActiveLegPosition == LegPosition.Four) offset = 6;
                if (ActiveLegPosition == LegPosition.Five) offset = 8;
                if (ActiveLegPosition == LegPosition.Six) offset = 10;


                waistCurrent.Text = t.ToString();
                dgDefaults.Rows[(int) ActiveLegPosition + offset].Cells[4].Value = t;
            });
        }

        private void hipPosition_Scroll(object sender, ScrollEventArgs e)
        {
            var t = e.NewValue; // + Convert.ToInt32(ssc.activeMin);

            ActiveJoint = ServoControl.Enums.Joint.Femur;


            _control.Move(ActiveLegPosition, ServoControl.Enums.Joint.Femur, t, 5000);

            var offset = 0;

            if (ActiveLegPosition == LegPosition.One) offset = 1;
            if (ActiveLegPosition == LegPosition.Two) offset = 3;
            if (ActiveLegPosition == LegPosition.Three) offset = 5;
            if (ActiveLegPosition == LegPosition.Four) offset = 7;
            if (ActiveLegPosition == LegPosition.Five) offset = 9;
            if (ActiveLegPosition == LegPosition.Six) offset = 11;

            Invoke((MethodInvoker) delegate
            {
                hipCurrent.Text = t.ToString();

                dgDefaults.Rows[(int) ActiveLegPosition + offset].Cells[4].Value = t;
            });
        }

        private void kneePosition_Scroll(object sender, ScrollEventArgs e)
        {
            var t = e.NewValue; // + Convert.ToInt32(ssc.activeMin);

            ActiveJoint = ServoControl.Enums.Joint.Tibia;


            _control.Move(ActiveLegPosition, ServoControl.Enums.Joint.Tibia, t, 5000);

            var offset = 0;

            if (ActiveLegPosition == LegPosition.One) offset = 2;
            if (ActiveLegPosition == LegPosition.Two) offset = 4;
            if (ActiveLegPosition == LegPosition.Three) offset = 6;
            if (ActiveLegPosition == LegPosition.Four) offset = 8;
            if (ActiveLegPosition == LegPosition.Five) offset = 10;
            if (ActiveLegPosition == LegPosition.Six) offset = 12;

            Invoke((MethodInvoker) delegate
            {
                kneeCurrent.Text = t.ToString();

                dgDefaults.Rows[(int) ActiveLegPosition + offset].Cells[4].Value = t;
            });
        }

        #endregion

        private void StartAutomation_Click(object sender, EventArgs e)
        {
            if (StartAutomation.Text.Contains("Start"))
            {
                _automation.Start();
                StartAutomation.Text = "Stop Robot";
            }
            else
            {
                _automation.Stop();
                StartAutomation.Text = "Start Robot";
            }
        }

        private void CenterTrigger_Scroll(object sender, ScrollEventArgs e)
        {
            _sensor.CenterThreshold = e.NewValue;
            CenterTriggerTxt.Text = e.NewValue.ToString();
        }

        private void ResetButton_Click(object sender, EventArgs e)
        {
            _sensor.ResetMap();
        }

        private void CenterTriggerTxt_TextChanged(object sender, EventArgs e)
        {

        }

        private void label18_Click(object sender, EventArgs e)
        {

        }

        private void BalanceChk_CheckedChanged(object sender, EventArgs e)
        {
            _inverseKinematics.BalanceMode = BalanceChk.Checked;
        }
    }
}